/*
 * arch/powerpc/platforms/mpc52xx_bestcomm.c
 *
 * Driver for MPC52xx processor BestComm peripheral controller
 * Using bplan GmbH OpenFirmware
 *
 * 2006 (c) bplan GmbH  This file is licensed under
 * the terms of the GNU General Public License version 2.  This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 *
 */

#undef DEBUG

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/module.h>

#include <asm/mpc52xx.h>
#include <asm/prom.h>

#include "bestcomm.h"

/**************/
/**************/
/**************/

struct bestcomm_mainhandle bestcomm_mainhandle;

static int opencnt = 0;
static struct bestcomm_taskhandle *mysdma_alloc(struct bestcomm_taskhandle *mytaskhandle, int queue_size);

/**************/
/**************/
/**************/

typedef u32 Cell;

/*
 * Virtutal <-> Bus address translator
 * everything is static inlien here
*/

/**************/
static void __iomem *
sram_bustovirt(struct bestcomm_mainhandle *mymainhandle, unsigned long busaddr)
{
        void __iomem *virtaddr;
        long offset;

        offset = busaddr - mymainhandle->sram_basebus;

        if (offset < 0)
                return NULL;

        if (offset > mymainhandle->sram_size)
                return NULL;

        virtaddr = (void __iomem   *) ( (u8 *)mymainhandle->sram_basevirt + offset);

        return virtaddr;
}

static unsigned long
sram_virttobus(struct bestcomm_mainhandle *mymainhandle, void *virtaddr)
{
        unsigned long busaddr;
        long offset;

        offset = (u8 *) virtaddr - (u8 *) mymainhandle->sram_basevirt;

        if (offset < 0)
                return 0;

        if (offset > mymainhandle->sram_size)
                return 0;

        busaddr = mymainhandle->sram_basebus + offset;

        return busaddr;
}

static void __iomem *
io_bustovirt(struct bestcomm_mainhandle *mymainhandle, unsigned long busaddr)
{
        void __iomem *virtaddr;
        long offset;

        offset = busaddr - mymainhandle->sdma_io_basebus;

        if (offset < 0)
                return NULL;

        if (offset > mymainhandle->sdma_io_size)
                return NULL;

        virtaddr = (void __iomem   *) ( (u8 *)mymainhandle->sdma_io_basevirt + offset);

        return virtaddr;
}

static unsigned long
io_virttobus(struct bestcomm_mainhandle *mymainhandle, void *virtaddr)
{
        unsigned long busaddr;
        long offset;

        offset = (u8 *) virtaddr - (u8 *) mymainhandle->sdma_io_basevirt;

        if (offset < 0)
                return 0;

        if (offset > mymainhandle->sdma_io_size)
                return 0;

        busaddr = mymainhandle->sdma_io_basebus + offset;

        return busaddr;
}

/**************/
static u8 *area1_end;
static u8 *area2_begin;
static spinlock_t sdma_lock = SPIN_LOCK_UNLOCKED;

void *
sdma_sram_alloc(int size, int alignment, u32 *dma_handle)
{
        u8 *a;

        spin_lock(&sdma_lock);

        /* alignment must be a power of 2 */
        BUG_ON(alignment & (alignment - 1));

        if (alignment < 16) {
                a = (u8 *)(((u32)area1_end + (alignment-1)) & ~(alignment-1));
                if (a + size <= area2_begin)
                        area1_end = a + size;
                else
                        a = 0;				/* out of memory */
        } else {
                a = (u8 *)(((u32)area2_begin - size) & ~(alignment - 1));
                if (a >= area1_end)
                        area2_begin = a;
                else
                        a = 0;				/* out of memory */
        }
        if(a && dma_handle)
                *dma_handle = sdma_sram_pa(a);
        spin_unlock(&sdma_lock);
        return (void *)a;
}

void __iomem*
bestcomm_sram_alloc(struct bestcomm_taskhandle *mytaskhandle, int len, unsigned long *busaddr)
{
        void *virtaddr;
        u32 lbusaddr=0;

        virtaddr =  sdma_sram_alloc(len, 32, &lbusaddr);
        *busaddr = (unsigned long) lbusaddr;

        return virtaddr;
}

/**************/
void
bestcomm_sram_free(struct bestcomm_taskhandle *mytaskhandle, void *ptr, int len)
{
}

/**************/
static int probemisc(struct bestcomm_mainhandle *mymainhandle)
{
        int interrupt;
        interrupt = 0xc0;
        mymainhandle->bestcomm_irq = interrupt;
        return 0;
}

/**************/
static int
probeavailable_sram(struct bestcomm_mainhandle *mymainhandle)
{
        struct device_node *sram_ofwnode;
        Cell *oneptr;
        int proplen;
        int nac;
        int nsc;
        int onepair_size;
        int paircount;
        int i;

        int pair_size;
        unsigned long pair_addr;

        int biggestpair_size=0;
        unsigned long biggestpair_addr=0;

        pr_debug("\n");

        sram_ofwnode = mymainhandle->sram_ofwnode;
        nac = prom_n_addr_cells(sram_ofwnode);
        nsc = prom_n_size_cells(sram_ofwnode);
        onepair_size = nac + nsc;

        oneptr = (Cell *) get_property(sram_ofwnode, "available", &proplen);
        if (!oneptr)
                return -1;

        paircount = proplen / (sizeof(Cell) * onepair_size);

        for(i=0; i < paircount; i++)
        {
                pair_addr = (unsigned long) oneptr[i * onepair_size];
                pair_size = (int) oneptr[i * onepair_size + nac];

                mymainhandle->sram_available += pair_size;

                pr_debug("SRAM free at 0x%8.8lx, size=%d\n", pair_addr, pair_size);

                if (pair_size>biggestpair_size) {
                        biggestpair_size = pair_size;
                        biggestpair_addr = pair_addr;
                }
        }

        area1_end = sram_bustovirt(mymainhandle, biggestpair_addr);
        area2_begin = sram_bustovirt(mymainhandle, biggestpair_addr + biggestpair_size);

        pr_debug("SRAM area %p -> %p\n", area1_end, area2_begin);

        return 0;
}

/**************/
static int
mapsram(struct bestcomm_mainhandle *mymainhandle)
{
        unsigned long sram_offset;
        int sram_size;
        void *__iomem sram_virtmem;

        sram_size = mymainhandle->sram_size;
        sram_offset = mymainhandle->sram_basebus;

        sram_virtmem = ioremap_nocache(sram_offset, sram_size);
        if (!sram_virtmem)
                return -1;

        mymainhandle->sram_basevirt = sram_virtmem;

        pr_debug("SRAM mapped at %p\n", sram_virtmem);

        return 0;
}

/**************/
static int
probesram(struct bestcomm_mainhandle *mymainhandle)
{
        struct device_node *onenode;
        Cell *oneptr;
        int proplen;
        int nac;
        int nsc;

        unsigned long sram_offset;
        int sram_size;

        pr_debug("\n");

	onenode = of_find_compatible_node(NULL, "memory", "mpc5200-sram");
        if (!onenode)
                return -1;

        nac = prom_n_addr_cells(onenode);
        nsc = prom_n_size_cells(onenode);

        pr_debug("nac=%d, nsc=%d\n", nac, nsc);

        oneptr = (Cell *) get_property(onenode, "reg", &proplen);
        if (!oneptr)
                return -1;

        if (proplen < 2)
                return -1;

        sram_offset = (unsigned long) oneptr[0];
        sram_size = (int) oneptr[nac];

        pr_debug("oneptr=%p. %lx %d\n", oneptr, sram_offset , sram_size);

        mymainhandle->sram_size = sram_size;
        mymainhandle->sram_basebus = sram_offset;
        mymainhandle->sram_ofwnode = onenode;

        return 0;
}

/**************/
static int
probetasktable(struct bestcomm_mainhandle *mymainhandle)
{
        struct device_node *onenode;
        Cell *oneptr;
        int proplen;
        int nac;
        unsigned long tdt_busaddr;
        long tdt_offset;
        int tdt_size;

        pr_debug("\n");

        onenode = mymainhandle->bestcomm_ofwnode;
        oneptr = (Cell *) get_property(onenode, "bestcomm_tasktable", &proplen);
        if (!oneptr)
                return -1;

        if (proplen < 2)
                return -1;

        nac = prom_n_addr_cells(onenode);

        tdt_busaddr = (unsigned long) oneptr[0];
        tdt_size = (int) oneptr[nac];

        tdt_offset = tdt_busaddr - mymainhandle->sram_basebus;

        pr_debug("tdt_busaddr=%8.8lx, tdt_size=%d, tdt_offset=%ld\n", tdt_busaddr, tdt_size, tdt_offset);

        if (tdt_offset < 0)
                return -1;

        if (tdt_offset > mymainhandle->sram_size)
                return -1;

        mymainhandle->sdma_tdtentry = (struct bestcomm_tdt __iomem *) ( (u8 *)mymainhandle->sram_basevirt + tdt_offset);

        pr_debug("SDMA_TDTEntry =%p\n", mymainhandle->sdma_tdtentry);

        return 0;
}

/**************/
static int
probeio(struct bestcomm_mainhandle *mymainhandle)
{
        struct device_node *onenode;
        Cell *oneptr;
        int proplen;
        int nac;
        int nsc;

        unsigned long io_busaddr;
        int io_size;

        pr_debug("\n");

        onenode = of_find_compatible_node(NULL, "dma-controller", "mpc5200-bestcomm");
        if (!onenode)
                return -1;

        nac = prom_n_addr_cells(onenode);
        nsc = prom_n_size_cells(onenode);

        pr_debug("nac=%d, nsc=%d\n", nac, nsc);

        oneptr = (Cell *) get_property(onenode, "reg", &proplen);
        if (!oneptr)
                return -1;

        if (proplen < 2)
                return -1;

        io_busaddr = (unsigned long) oneptr[0];
        io_size = (int) oneptr[nac];

        pr_debug("oneptr=%p. %lx %d\n", oneptr, io_busaddr , io_size);
        mymainhandle->sdma_io_size = io_size;
        mymainhandle->sdma_io_basebus = io_busaddr;
        mymainhandle->bestcomm_ofwnode = onenode;

        return 0;
}

/**************/
static int
mapio(struct bestcomm_mainhandle *mymainhandle)
{
        unsigned long io_busaddr;
        void __iomem *io_virtaddr;
        int io_size;

        io_busaddr = mymainhandle->sdma_io_basebus;
        io_size = mymainhandle->sdma_io_size;

        io_virtaddr = ioremap_nocache(io_busaddr, io_size);
        if (!io_virtaddr)
                goto fail;

        mymainhandle->sdma_io_basevirt = io_virtaddr;

        pr_debug("Bestcomm mapped at %p\n", io_virtaddr );

        return 0;

fail:
        return -1;
}


/**************/
static int
do_bestcomminit(struct bestcomm_mainhandle *mymainhandle)
{
        int ret;

        ret = probeio(mymainhandle);
        if (ret<0)
                goto fail;

        ret = mapio(mymainhandle);
        if (ret<0)
                goto fail;

        ret = probesram(mymainhandle);
        if (ret<0)
                goto fail;

        ret = mapsram(mymainhandle);
        if (ret<0)
                goto fail;

        ret = probeavailable_sram(mymainhandle);
        if (ret<0)
                goto fail;

        ret = probetasktable(mymainhandle);
        if (ret<0)
                goto fail;

        ret = probemisc(mymainhandle);
        if (ret<0)
                goto fail;

        return 0;

fail:
        return -1;
}

/**************/
int
bestcomm_hwiscapable(void)
{
        return bestcomm_mainhandle.buildin_ofwnode != NULL;
}

/**************/
/*
 * Entry, should be called when the CPU is running
 * on an OFW machine.
 * We will silently and properly failed if no compatible
 * hardware (MPC5200, MPC5200b yet)
*/
int
bestcomm_init_once(void)
{
        int ret;

        pr_debug("opencnt %d\n", opencnt);

        /* Already inited ? */
        if (opencnt>0)
                return 0;

        memset(&bestcomm_mainhandle, 0x00, sizeof(bestcomm_mainhandle) );

        ret = do_bestcomminit(&bestcomm_mainhandle);
        if (ret<0)
                return -1;

        opencnt++;

        printk(KERN_INFO "MPC52xx/OpenFirmware: %d kB of free SRAM\n", bestcomm_mainhandle.sram_available );

        return 0;
}

/**************/
/**************/
/**************/
void
bestcomm_taskfree(struct bestcomm_taskhandle *mytaskhandle)
{
        if (mytaskhandle)
        {
                kfree(mytaskhandle->cookie);
                mytaskhandle->cookie = NULL;
        }
}

struct bestcomm_taskhandle *
bestcomm_taskallocate(int bestcomm_tasknum, int queue_size)
{
        struct bestcomm_taskhandle *mytaskhandle;
        struct bestcomm_tdt __iomem   *bestcomm_tdt;

        if (bestcomm_tasknum<0)
                return NULL;

        if (bestcomm_tasknum>=BESTCOMM_MAX_TASKS)
                return NULL;

        mytaskhandle = &bestcomm_mainhandle.taskhandle_table[bestcomm_tasknum];

        pr_debug("bestcomm_tasknum %d\n", bestcomm_tasknum);

        bestcomm_tdt = &bestcomm_mainhandle.sdma_tdtentry[bestcomm_tasknum];

        mytaskhandle->taskid = bestcomm_tasknum;
        mytaskhandle->bestcomm_taskcode = sram_bustovirt(&bestcomm_mainhandle, bestcomm_tdt->start);
        mytaskhandle->bestcomm_vartable = sram_bustovirt(&bestcomm_mainhandle, bestcomm_tdt->var);
        mytaskhandle->irq = bestcomm_mainhandle.bestcomm_irq + bestcomm_tasknum;

	pr_debug("irq=%d\n", mytaskhandle->irq );

        mytaskhandle = mysdma_alloc(mytaskhandle, queue_size);

        return mytaskhandle;
}

/**************/
int
bestcomm_taskenable(struct bestcomm_taskhandle *mytaskhandle)
{
        int task;
        u16 reg;

        task = mytaskhandle->taskid;

        reg = in_be16(&bestcomm_mainhandle.sdma_io_basevirt->tcr[task]);
        out_be16(&bestcomm_mainhandle.sdma_io_basevirt->tcr[task], reg | 0x8000);

        return task;
}

/**************/
int
bestcomm_taskdisable(struct bestcomm_taskhandle *mytaskhandle)
{
        int task;
        u16 reg;

        task = mytaskhandle->taskid;

        reg = in_be16(&bestcomm_mainhandle.sdma_io_basevirt->tcr[task]);
        out_be16(&bestcomm_mainhandle.sdma_io_basevirt->tcr[task], reg & ~0x8000);

        return task;
}

/**************/
int
bestcomm_taskclear_irq(struct bestcomm_taskhandle *mytaskhandle)
{
        int tasknum =  mytaskhandle->taskid;
        out_be32(&bestcomm_mainhandle.sdma_io_basevirt->IntPend, 1 << tasknum);
        return tasknum;
}

/**************/
/**************/
/**************/


/**************/
void __iomem *
bestcomm_phys_to_virt(struct bestcomm_mainhandle *mymainhandle, unsigned long busaddr)
{
        return sram_bustovirt(mymainhandle, busaddr);
}

unsigned long
bestcomm_virt_to_phys(struct bestcomm_mainhandle *mymainhandle, void __iomem* virtaddr)
{
        return sram_virttobus(mymainhandle, virtaddr);
}

/**************/
/**************/
/**************/

/*
 * friendly helper func
*/

int
bestcomm_taskget_irq(struct bestcomm_taskhandle *mytaskhandle)
{
        return mytaskhandle->irq;
}

u32 __iomem*
bestcomm_taskget_code(struct bestcomm_taskhandle *mytaskhandle)
{
        return mytaskhandle->bestcomm_taskcode;
}

u32 __iomem*
bestcomm_taskget_vartable(struct bestcomm_taskhandle *mytaskhandle)
{
        return mytaskhandle->bestcomm_vartable;
}

u32 __iomem*
bestcomm_taskget_inctable(struct bestcomm_taskhandle *mytaskhandle)
{
        return mytaskhandle->bestcomm_vartable + BESTCOMM_MAX_VAR;
}

u16 __iomem*
bestcomm_taskget_tcr(struct bestcomm_taskhandle *mytaskhandle)
{
        return &bestcomm_mainhandle.sdma_io_basevirt->tcr[mytaskhandle->taskid];
}


/**************/
/**************/
/**************/
/*
 * Stuff copied and a little bit modified from the original Linux code
 * indeed, I think it makes sense to reuse code which works. Moreover,
 * it's then easier to reuse existing driver with SDMA task which are
 * not in the OFW
*/

/**************/
unsigned long
sdma_sram_pa(void __iomem *virt)
{
        return sram_virttobus(&bestcomm_mainhandle, virt);
}

void __iomem *
sdma_sram_va(unsigned long pa)
{
        return sram_bustovirt(&bestcomm_mainhandle, pa);
}

unsigned long
sdma_io_pa(void __iomem *virt)
{
        return io_virttobus(&bestcomm_mainhandle, virt);
}

void __iomem *
sdma_io_va(unsigned long pa)
{
        return io_bustovirt(&bestcomm_mainhandle, pa);
}

/**************/
/*
 * Here, I just alloc and setup the buffer management stuff
 * for this task
*/
static struct bestcomm_taskhandle *
mysdma_alloc(struct bestcomm_taskhandle *mytaskhandle, int queue_size)
{
        void **cookie;

        if (!mytaskhandle)
                return NULL;

        if (queue_size)
        {
                cookie = kmalloc(sizeof(*cookie) * queue_size, GFP_KERNEL);
                if (!cookie) {
                        return NULL;
                }

                mytaskhandle->cookie = cookie;
        }

        mytaskhandle->num_bd = queue_size;
        return mytaskhandle;
}

/**************/
/*
 * Export symbol for modules
*/

EXPORT_SYMBOL(bestcomm_init_once);
EXPORT_SYMBOL(bestcomm_hwiscapable);
EXPORT_SYMBOL(bestcomm_taskallocate);
EXPORT_SYMBOL(bestcomm_taskfree);
EXPORT_SYMBOL(bestcomm_taskclear_irq);
EXPORT_SYMBOL(bestcomm_taskenable);
EXPORT_SYMBOL(bestcomm_taskdisable);
EXPORT_SYMBOL(bestcomm_taskget_irq);
EXPORT_SYMBOL(bestcomm_taskget_code);
EXPORT_SYMBOL(bestcomm_taskget_vartable);
EXPORT_SYMBOL(bestcomm_taskget_inctable);
EXPORT_SYMBOL(bestcomm_taskget_tcr);
EXPORT_SYMBOL(bestcomm_phys_to_virt);
EXPORT_SYMBOL(bestcomm_virt_to_phys);
EXPORT_SYMBOL(bestcomm_sram_alloc);
EXPORT_SYMBOL(bestcomm_sram_free);
