/*
 *  mpc52xx.c
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */

/*
 * Here we "translate" Linux API mess to our funcs
*/

#include <linux/major.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/blkdev.h>
#include <linux/bitops.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/fs.h>

/**************/
/**************/
/**************/
#include "mpc52xx_blockata.h"

/**************/
/**************/
/**************/
static int local_setpiomode(struct mpc52xx_blockata_priv *priv, int piomode_arg);

/**************/
/**************/
/**************/

static void dump_config(struct mpc52xx_blockata_priv *priv)
{
        VPRINTK(" configuration:\n");

        if (priv->multi_available)
                VPRINTK("\t * Multi-PIO commands: available (%d sectors per shot)\n", priv->multi_secpercmd );
        VPRINTK("\t * PIO Mode: %d\n", priv->piomode < 0 ? 0 : priv->piomode);

        if (priv->UsePacket)
                VPRINTK("\t * Packet feature set supported\n");

        if (priv->IsATAPI)
                VPRINTK("\t * %s ATAPI device detected\n", priv->IsRemovable ? "Removable" : "");

        VPRINTK("\t * LBA48 supported: %s\n", priv->drive_canlba48 ? "Yes" : "No");
        VPRINTK("\t * CHS: %s (%d/%d/%d)\n", priv->drive_chs_ok ? "Ok" : "Ko", priv->drive_chs_cylinders, priv->drive_chs_heads , priv->drive_chs_sectorspertrack);
        VPRINTK("\t * SDMA Engine: %s\n",
#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
                "Enable");
#else
                "Disable"
               );
#endif
}

/**************/
/**************/
/**************/



/* ATAPI-4 PIO specs (arranged for the 5200, cfr User Manual) */
/* numbers in ns, extrapolation done by code */
static int ataspec_t0[5]    = {600, 383, 240, 180, 120};
static int ataspec_t1[5]    = { 70,  50,  30,  30,  25};
static int ataspec_t2_8[5]  = {290, 290, 290,  80,  70};
static int ataspec_t2_16[5] = {165, 125, 100,  80,  70};
static int ataspec_t2i[5]   = {  0,   0,   0,  70,  25};
static int ataspec_t4[5]    = { 30,  20,  15,  10,  10};
static int ataspec_ta[5]    = { 35,  35,  35,  35,  35};



/**************/
/**************/
/**************/
static int device_handlebio(
        struct mpc52xx_blockata_priv *priv,
        struct request *req,
        struct bio *bio,
        int *sectorcnt_ptr)
{
        int i;
        struct bio_vec *bvec;
        sector_t sector;
        int sectorcnt_biovec;
        int sectorcnt;
        int is_write;
        int ret;

        ret = -1;
        is_write = bio_data_dir(bio);
        sector = bio_getsector(bio);

        sectorcnt = 0;

        NPRINTK("bio=%p\n", bio);

        bio_for_each_segment(bvec, bio, i)
        {
                char *buffer;

                sectorcnt_biovec = bio_cur_sectors_idx(bio, i);
                buffer = __bio_kmap_atomic(bio, i, KM_USER0);
                ret = mpc52xx_ata_dotransfer(
                              priv,
                              req, bio, i,
                              sector, sectorcnt_biovec,
                              buffer, is_write);

                if (ret==0)
                        goto end;

                // Success or error, we kunmap the buffer -> it's done!
                __bio_kunmap_atomic(buffer, KM_USER0);

                if (ret<0)
                        goto end;

                /* Ret > 0 -> done, do the next one */
                sector += ret;
                sectorcnt += ret;
        }

        ret = sector;

end:
        *sectorcnt_ptr = sectorcnt;
        return ret;
}

/**************/
/**************/
/**************/
/*
 * The main challenge!
*/

static int device_handlereq(
        struct mpc52xx_blockata_priv *priv,
        struct request *req,
        int *sectorcnt_ptr)
{
        struct bio *bio;
        int ret;
        int nsect;
        int nsect_thisbio;

        ret = 0;
        nsect = 0;

        if (priv->UsePacket)
                return -ENOTSUPP;

        if (priv->io_inprogress)
        {
                VPRINTK("IO already in-progress, can not do more!\n");
                ret = -EBUSY;
                goto end;
        }


        rq_for_each_bio(bio, req)
        {
                ret = device_handlebio(priv, req, bio, &nsect_thisbio);

                if (ret<=0)
                        goto end;

                nsect += nsect_thisbio;
        }

        ret = nsect;

end:
        *sectorcnt_ptr = nsect;
        return ret;
}

/**************/
static void device_request(request_queue_t *q)
{
        struct mpc52xx_blockata_priv *priv = (struct mpc52xx_blockata_priv *) q->queuedata;
        struct request *req;
        int ret=0;
        int sectorcnt;

        while ((req = elv_next_request(q)) != NULL) {

                if (! blk_fs_request(req)) {
                        printk (KERN_NOTICE "Skip non-fs request\n");
                        end_request(req, 0);
                        continue;
                }

                ret = device_handlereq(priv, req, &sectorcnt);
                if (ret==0) {
                        NPRINTK("stop the queue\n");
                        blk_stop_queue(q);
                        break;
                } else if (ret>0)
                        end_request(req, 1);
                else
                        end_request(req, 0);
        }

        NPRINTK("end, ret=%d\n", ret);
}


/**************/
/**************/
/**************/

static int device_doreset(
        struct mpc52xx_blockata_priv *priv,
        int do_drivereset)
{
        int ret = 0;

        NPRINTK("priv->io_inprogress = %d, do_drivereset=%d\n",
                priv->io_inprogress, do_drivereset );

        if (priv->io_inprogress != 0)
                mpc52xx_ata_ack_blkreq(priv, 0);

        ret = mpc52xx_ata_doreset(priv);
        if(ret<0)
                ret = -EIO;

        return ret;
}

/*
 * The device open/close func
*/

/**************/
static int device_open(struct inode *inode, struct file *filp)
{
        unsigned long flags;
        struct mpc52xx_blockata_priv *priv = inode->i_bdev->bd_disk->private_data;
        int ret = 0;
        int force_enableirq;

        NPRINTK("enter. sectors=%lld, usercnt=%d\n", priv->drive_sectors, priv->usercnt);

        spin_lock_irqsave(&priv->lock, flags);
        force_enableirq = 0;

        if (priv->usercnt >= DEVICE_USERLIMIT)
        {
                VPRINTK("Error! Too much openers (%d)!\n", priv->usercnt);
                ret = -1;
                goto end;
        }

        // device not inited ? -> try to init it a gain
        if (!priv->drive_inited)
        {
                priv->drive_sectors = 0;
                priv->drive_inited = 0;

                ret = mpc52xx_ata_init_drive(priv);
                if (ret>=0)
                        priv->drive_inited = 1;

                ret = 0;
        }

        NPRINTK("drive inited ok. s=%x, ret=%d\n", read_altstatus(priv), ret);
        NPRINTK("drive looks ok. s=%x\n", read_altstatus(priv));

        if ( (priv->usercnt==0) || force_enableirq )
        {
                priv->ata_handler = ata_void_handler;
                priv->sdma_handler = sdma_void_handler;

                write_ctl(priv, 0);

                priv->curio_bio = NULL;
                priv->curio_req = NULL;
                priv->io_inprogress = 0;
        }

        priv->usercnt++;

end:
        spin_unlock_irqrestore(&priv->lock, flags);

        NPRINTK("end. ret=%d\n", ret);
        return ret;
}

/**************/
static int device_release(struct inode *inode, struct file *filp)
{
        unsigned long flags;
        int ret;
        struct mpc52xx_blockata_priv *priv = inode->i_bdev->bd_disk->private_data;

        ret = 0;

        NPRINTK("enter usercnt=%d\n", priv->usercnt);

        spin_lock_irqsave(&priv->lock, flags);
        priv->usercnt--;

        if (priv->usercnt==0)
        {
                write_ctl(priv, ATA_NIEN);

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
                sdma_ata_disable(priv->sdma);
                sdma_ata_clear_irq(priv->sdma);
                sdma_ata_reset(priv->sdma);
#endif
        }

        if (priv->usercnt<0)
                VPRINTK("Warning! _release and _open count doesn't match!\n");

        NPRINTK("end, ret=%d\n", ret);

        spin_unlock_irqrestore(&priv->lock, flags);

        return ret;
}

/**************/
/**************/
/**************/
/*
 * The device ioctl
*/

#ifndef HDIO_GETGEO_BIG

/*
 * Let's define GETGEO big because it's used!
 * they removed it from recent linux kernel
 * but already compiled software use it (eg: hdparm)
 *
 * this define and struct has been picked from linux 2.4.32
*/

#define HDIO_GETGEO_BIG 0x330
struct hd_big_geometry
{
        unsigned char heads;
        unsigned char sectors;
        unsigned int cylinders;
        unsigned long start;
};
#endif

static int device_ioctl(
        struct inode *inode,
        struct file *filp,
        unsigned int cmd,
        unsigned long arg)
{
        int ret;
        unsigned long flags;
        struct mpc52xx_blockata_priv *priv = inode->i_bdev->bd_disk->private_data;

        NPRINTK("%s: cmd=0x%x\n", __func__, cmd);

        spin_lock_irqsave(&priv->lock, flags);

        switch(cmd)
        {
        case HDIO_SET_MULTCOUNT: {
                u16 multicount;

                NPRINTK("HDIO_SET_PIO_MODE. Try to set the multicount value to %ld\n", arg);

                if (priv->UsePacket) {
                        ret = -EIO;
                        break;
                }

                multicount = (u16) arg;

                ret = mpx52xx_ata_dosetmulti(priv, multicount);
                if (ret<0)
                        ret = -EFAULT;
        }
        break;

        case HDIO_GET_MULTCOUNT:
                if (priv->UsePacket) {
                        ret = -EIO;
                        break;
                }

                ret = put_user(priv->multi_secpercmd , (long __user *) arg);
                break;

        case HDIO_DRIVE_CMD: {
                u8 drivecmd_args[4+512];
                u8 drivecmd;
                int lentocopy;

                if (NULL == (void *)arg) {
                        ret = -EINVAL;
                        break;
                }

                if (copy_from_user(drivecmd_args, (void __user *) arg, sizeof(drivecmd_args) )) {
                        ret = -EFAULT;
                        break;
                }

                drivecmd = drivecmd_args[0];
                ret = mpc52xx_ata_dodrivecmd(priv, &flags, drivecmd_args);

                lentocopy =
                        ( (drivecmd==ATA_CMD_ID_ATAPI) || (drivecmd==ATA_CMD_ID_ATA) )
                        ? (512 + 4): 4;

                if (copy_to_user((void __user *)arg, drivecmd_args, lentocopy))
                        ret = -EFAULT;
        }
        break;

        case HDIO_SET_PIO_MODE: {
                int piomode;
                NPRINTK("HDIO_SET_PIO_MODE. Try to set PIO %ld\n", arg);

                if (priv->io_inprogress) {
                        VPRINTK("IO already in-progress, can not do more!\n");
                        ret = -EBUSY;
                        break;
                }

                priv->io_inprogress = 1;
                piomode = local_setpiomode(priv, (int) arg);
                priv->io_inprogress = 0;

                if (piomode>0)
                        VPRINTK("PIO mode %d sat\n", piomode);

                ret = 0;
        }
        break;

        case HDIO_GET_IDENTITY: {
                NPRINTK("HDIO_GET_IDENTITY:\n" );

                if (priv->drive_identify_valid) {
                        if (copy_to_user((void __user *) arg, priv->drive_identify, sizeof(priv->drive_identify) ))
                                ret = -EFAULT;
                        else
                                ret = 0;
                } else
                        ret = -1;
        }
        break;

        case HDIO_GETGEO: {
                struct hd_geometry geo;

                NPRINTK("HDIO_GETGEO\n");

                geo.cylinders = priv->drive_cylinders;
                geo.heads = priv->drive_heads;
                geo.sectors = priv->drive_sectorspertrack;
                geo.start = 0;
                if (copy_to_user((void __user *) arg, &geo, sizeof(geo)))
                        ret = -EFAULT;
                else
                        ret = 0;

                NPRINTK("HDIO_GETGEO, ret=%d\n", ret);
        }
        break;

        case HDIO_GETGEO_BIG: {
                struct hd_big_geometry biggeo;

                NPRINTK("HDIO_GETGEO_BIG\n");

                biggeo.cylinders = priv->drive_cylinders;
                biggeo.heads = priv->drive_heads;
                biggeo.sectors = priv->drive_sectorspertrack;
                biggeo.start = 0;

                if (copy_to_user((void __user *) arg, &biggeo, sizeof(biggeo)))
                        ret = -EFAULT;
                else
                        ret = 0;

                NPRINTK("HDIO_GETGEO_BIG, ret=%d\n", ret);
        }
        break;

        case HDIO_DRIVE_RESET: {
                VPRINTK("Issue a controller and drive reset\n");
                ret = device_doreset(priv, 1);
                if(ret<0)
                        ret = -EIO;

                break;
        }

        default:
                ret = -EINVAL;
        }

        NPRINTK("%s: end, ret=%d\n", __func__, ret);

        spin_unlock_irqrestore(&priv->lock, flags);

        return ret;
}

/**************/
/**************/
/**************/
/*
 * The device media changes/revalidate
*/
static int device_media_changed(struct gendisk *gd)
{
        /* Our media won't move! */
        NPRINTK("\n");
        return 0;
}

static int device_revalidate_disk(struct gendisk *gd)
{
        NPRINTK("\n");
        return 0;
}

/**************/
/**************/
/**************/
/*
 * Modules prove/init
*/

static struct block_device_operations device_fops =
        {
                .owner   = THIS_MODULE,
                           .open	   = device_open,
                                      .release = device_release,
                                                 .ioctl   = device_ioctl,
                                                            .media_changed = device_media_changed,
                                                                             .revalidate_disk = device_revalidate_disk,
                                                                                        };

/**************/
static void module_free(struct mpc52xx_blockata_priv *priv)
{
        NPRINTK("Enter. priv=%p\n", priv);
        if (priv)
        {
                NPRINTK("free private private\n");

		if (priv->ata_irq>=0)
			free_irq(priv->ata_irq, priv);

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
		if (priv->sdma_irq>=0)
			free_irq(priv->sdma_irq, priv);
#endif

                NPRINTK("priv->device_queue=%p\n", priv->device_queue);
                if (priv->device_queue)
                        blk_cleanup_queue(priv->device_queue);

                NPRINTK("priv->device_gendisk=%p\n", priv->device_gendisk);
                if (priv->device_gendisk)
                        del_gendisk(priv->device_gendisk);

                NPRINTK("priv->major=%d\n", priv->major);
                if (priv->major>0)
                        unregister_blkdev(priv->major, DEVSKEL_DEVICENAME);

                NPRINTK("free priv. p=%p\n", priv);
                kfree(priv);
        }

        NPRINTK("End\n");
}

static void module_remove(struct mpc52xx_blockata_priv *priv)
{
        printk(KERN_INFO DEVSKEL_DRIVERNAME ": Tchuss!\n");
        module_free(priv);
}



/**************/
/**************/
/**************/

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA

irqreturn_t generalsdma_handler(int irq, void *host)
         {
                 struct mpc52xx_blockata_priv *priv = (struct mpc52xx_blockata_priv *) host;
                 irqreturn_t ret = IRQ_NONE;

                 NPRINTK2("%s: irq=%d, sdma_handler %p\n", __func__, irq, priv->sdma_handler);

                 if (irq == priv->sdma_irq)
                 {
                         ret =
                                 (priv->sdma_handler)
                                 ? priv->sdma_handler(priv)
                                 : IRQ_HANDLED;
                 } else
                 {
                         ret = IRQ_NONE;
                 }

                 return ret;
         }


#endif

         irqreturn_t generalata_handler(int irq, void *host)
                  {
                          struct mpc52xx_blockata_priv *priv = (struct mpc52xx_blockata_priv  *) host;
                          irqreturn_t ret;

                          if (irq == priv->ata_irq)
                          {
                                  u8 status;
                                  status = read_status(priv);

                                  ret =
                                          (priv->ata_handler)
                                          ? priv->ata_handler(priv, status)
                                          : IRQ_HANDLED;
                          } else
                          {
                                  ret = IRQ_NONE;
                          }

                          return ret;
                  }


                  static void
                  mpc52xx_ide_apply_timing(struct mpc52xx_ata __iomem *regs, struct mpc52xx_ata_timings *timing)
                  {
                          out_be32(&regs->pio1,  timing->pio1);
                          out_be32(&regs->pio2,  timing->pio2);
                          out_be32(&regs->mdma1, timing->mdma1);
                          out_be32(&regs->mdma2, timing->mdma2);
                          out_be32(&regs->udma1, timing->udma1);
                          out_be32(&regs->udma2, timing->udma2);
                          out_be32(&regs->udma3, timing->udma3);
                          out_be32(&regs->udma4, timing->udma4);
                          out_be32(&regs->udma5, timing->udma5);
                  }

                  static void
                  mpc52xx_ide_compute_pio_timing( struct mpc52xx_ata_timings *timing, unsigned int ipb_period, u8 pio)
                  {
                          u32 t0, t2_8, t2_16, t2i, t4, t1, ta;

                          /* We add 1 as a 'margin' */
                          t0    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t0[pio]);
                          t2_8  = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t2_8[pio]);
                          t2_16 = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t2_16[pio]);
                          t2i   = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t2i[pio]);
                          t4    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t4[pio]);
                          t1    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_t1[pio]);
                          ta    = 1 + CALC_CLK_VALUE_UP(ipb_period, 1000*ataspec_ta[pio]);

                          timing->pio1 = (t0 << 24) | (t2_8 << 16) | (t2_16 << 8) | (t2i);
                          timing->pio2 = (t4 << 24) | (t1 << 16) | (ta << 8);
                  }

                  static int local_setpiomode(struct mpc52xx_blockata_priv *priv, int piomode_arg)
                  {
                          int piomode;
                          int ret;

                          NPRINTK(KERN_DEBUG DEVSKEL_DEVICENAME "pio mode arg=%d\n", piomode_arg);

                          piomode = mpc52xx_ata_isthispiovalid(priv, piomode_arg);
                          if (piomode<0)
                                  return -1;

                          NPRINTK(KERN_DEBUG DEVSKEL_DEVICENAME "pio mode=%d\n", piomode);

                          ret = mpc52xx_ata_setpiomode(priv, piomode);
                          if (ret<0)
                                  return -1;

                          mpc52xx_ide_compute_pio_timing(&priv->timings[0], priv->ipb_period, piomode);
                          mpc52xx_ide_apply_timing(priv->ata_regs, &priv->timings[0]);

                          return piomode;
                  }

                  static int
                  mpc52xx_ide_setup(
                          struct mpc52xx_ata __iomem *regs,
                          struct mpc52xx_blockata_priv *priv)
                  {

#define MPC52xx_IPBFREQ (132*1000*1000)

                          /* Vars */

                          int tslot;

                          NPRINTK("%s: enter\n", __func__);

                          out_8(&regs->dma_mode, MPC52xx_ATA_DMAMODE_FR);
                          udelay(10);

                          /* All sample code do this */
                          out_be32(&regs->share_cnt, 0);

                          /* Configure & Reset host */
                          out_be32(&regs->config,
                                   MPC52xx_ATA_HOSTCONF_IE |
                                   MPC52xx_ATA_HOSTCONF_IORDY |
                                   MPC52xx_ATA_HOSTCONF_SMR |
                                   MPC52xx_ATA_HOSTCONF_FR);
                          udelay(10);
                          out_be32(&regs->config,
                                   MPC52xx_ATA_HOSTCONF_IE |
                                   MPC52xx_ATA_HOSTCONF_IORDY);

                          /* Get IPB bus period */
                          priv->ipb_period = 1000000000 / (MPC52xx_IPBFREQ/1000);

                          /* Try to set the time slot to around 1us = 1000000 ps */
                          tslot = CALC_CLK_VALUE_UP(priv->ipb_period, 1000000);
                          out_be32(&regs->share_cnt, tslot << 16);


                          /* Init imings to PIO0 (safest) */
                          memset(priv->timings, 0x00, 2*sizeof(struct mpc52xx_ata_timings));

                          mpc52xx_ide_compute_pio_timing(&priv->timings[0], priv->ipb_period, 0);
                          mpc52xx_ide_compute_pio_timing(&priv->timings[1], priv->ipb_period, 0);

                          mpc52xx_ide_apply_timing(regs, &priv->timings[0]);

                          return 0;
                  }




                  static int some_hwinit(struct mpc52xx_blockata_priv *priv)
                  {
                          struct mpc52xx_gpio __iomem *gpio_regs;
                          struct mpc52xx_ata __iomem *ata_regs;
			struct device_node *of_dev;	
                          int ata_irq;
                          u32 res_mem;
                          int sdma_irqnum;
                          int ret;

                          ata_irq = -1;
                          sdma_irqnum = -1;

                          ata_regs = NULL;
                          gpio_regs = NULL;
                          res_mem = 0;



                          priv->ata_irq = -1;
                          priv->sdma_irq = -1;

			of_dev = of_find_compatible_node(NULL, "ata", "mpc5200-ata");
			if (of_dev == NULL)
				return -ENODEV;

		       
                          /* Get the resources of this device */
                          ata_irq = irq_of_parse_and_map(of_dev, 0);
                          if (ata_irq<0)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Invalid resource set!\n");
                                  return -EINVAL;
                          }

                          ret = request_irq(ata_irq, generalata_handler, SA_INTERRUPT, DEVSKEL_DRIVERNAME " ATA interrupt", priv);
                          if (ret)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can not allocate interrupt for the ATA controller\n");
                                  ata_irq=-1;
                                  goto error;
                          }
                          priv->ata_irq = ata_irq;

                          res_mem = mpc52xx_ata_getregisterbase(priv);
                          if (!res_mem)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Unable to locate ATA registers\n");
                                  ret = -ENOMEM;
                                  goto error;
                          }

                          ata_regs = ioremap(res_mem, sizeof(struct mpc52xx_ata));
                          if (!ata_regs)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Unable to ioremap ATA registers\n");
                                  ret = -ENOMEM;
                                  goto error;
                          }

                          priv->ata_regs = ata_regs;
                          priv->ata_regs_bus = res_mem;

                          /* Setup the ATA controller */
                          ret = mpc52xx_ide_setup(ata_regs, priv);
                          if (ret)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Controller setup failed !\n");
                                  goto error;
                          }

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
                          ret = mpc52xx_ata_sdma_setup(priv);
                          if (ret)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": SDMA setup failed !\n");
                                  goto error;
                          }

                          sdma_irqnum = sdma_ata_getirq(priv->sdma);
                          ret = request_irq(sdma_irqnum, generalsdma_handler, SA_INTERRUPT, DEVSKEL_DRIVERNAME " SDMA interrupt", priv);
                          if (ret)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can not allocate interrupt for the SDMA Task\n");
                                  goto error;
                          }

                          priv->sdma_irq = sdma_irqnum;

                          NPRINTK("%s: ATA irq=%d, SDMA IRQ=%d\n", __func__, ata_irq, priv->sdma_irq);

#endif

                          return 0;

                  error:
                          return ret;
                  }

                  /**************/
                  static struct mpc52xx_blockata_priv *module_probe(int *retcode)
                  {
                          int ret;
                          struct mpc52xx_blockata_priv *priv;
                          struct gendisk *device_gendisk;
                          struct request_queue *device_queue;

                          ret = 0;
                          priv = NULL;
                          device_gendisk = NULL;
                          device_queue = NULL;

                          NPRINTK("enter\n");

								
                          /* Setup private structure */
                          priv = kmalloc(sizeof(*priv), GFP_ATOMIC);
                          if (!priv)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can't allocate private structure !\n");
                                  ret = -ENOMEM;
                                  goto error;
                          }
                          memset(priv, 0, sizeof (*priv) );


                          spin_lock_init(&priv->lock);
                          priv->major = -1;

                          NPRINTK("device privata data ok. p=%p\n", priv);

                          ret = some_hwinit(priv);
                          if(ret<0)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can't allcoate some private stuff !\n");
                                  goto error;
                          }

                          priv->drive_sectors = 0;
                          priv->drive_inited = 0;

			  /*
			   * I prefer to the drive init here. Indeed, in case of failure
			   * (for example no drive present), the linux block code generally exploses
			   * 
			  */
                          ret = mpc52xx_ata_init_drive(priv);
                          if (ret<0)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can not init the ATA drive\n");
                                  goto error;
                          }

                          ret = register_blkdev(0, DEVSKEL_DEVICENAME);
                          if (ret<0)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can not register the block device\n");
                                  goto error;
                          }
                          priv->major = ret;

                          NPRINTK("block device registered with major %d\n", ret);

                          // minors must be >1 for partition mess (?)
                          device_gendisk = alloc_disk(32);
                          if (!device_gendisk)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can not allocate a block disk\n");
                                  goto error;
                          }
                          priv->device_gendisk = device_gendisk;

                          device_gendisk->major = priv->major;
                          device_gendisk->first_minor = 0;
                          device_gendisk->fops = &device_fops;
                          sprintf(device_gendisk->disk_name, DEVSKEL_DEVICENAME);

                          NPRINTK("device disk allocated. p=%p\n", device_gendisk);

                          device_queue = blk_init_queue(device_request, &priv->lock);
                          if (!device_queue)
                          {
                                  printk(KERN_ERR DEVSKEL_DRIVERNAME ": Can not init the block queue\n");
                                  ret = -1;
                                  goto error;
                          }

                          priv->device_queue = device_queue;
                          device_gendisk->queue = device_queue;

                          NPRINTK("block queue ok. p=%p\n", device_queue);

                          device_queue->queuedata = priv;
                          device_gendisk->private_data = priv;


#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_MAXPIO
                          NPRINTK("Set pio mode to max\n");
                          priv->piomode = local_setpiomode( priv, -1);
#endif

                          priv->drive_inited = 1;

                          /* Only add the disk as very last step */
                          blk_queue_max_phys_segments(device_queue, DRIVER_MAXHWSEG);
                          blk_queue_max_sectors(priv->device_queue, DRIVER_MAXHWSEG);

                          // This should be the multi pio value
                          blk_queue_max_hw_segments(device_queue, priv->multi_secpercmd);

                          set_capacity(priv->device_gendisk, priv->drive_sectors);
                          add_disk(device_gendisk);

                          init_waitqueue_head(&priv->my_waitqueue);

                          printk(KERN_INFO DEVSKEL_DRIVERNAME ": %s (Version %s - Compiled date %s at %s)\n", DEVSKEL_DRIVERNAME, DEVSKEL_DRIVERVERSION, __DATE__, __TIME__);
                          dump_config(priv);

                          *retcode = 0;
                          return priv;

                  error:

                          module_free(priv);

                          *retcode = (ret == 0) ? -1 : ret;
                          return NULL;
                  }

                  /**************/
                  /**************/
                  /**************/
                  /*
                   * Kernel call for module load/remove
                  */
                  struct mpc52xx_blockata_priv *global_priv = NULL;

static int __init kernelcall_init(void)
{
        int retcode;

        global_priv = module_probe(&retcode);
        return retcode;
}

static void __exit kernelcall_exit(void)
{
        module_remove(global_priv);
        global_priv = NULL;
}


module_init(kernelcall_init);
module_exit(kernelcall_exit);

MODULE_AUTHOR("bplan GmbH");
MODULE_DESCRIPTION(DEVSKEL_DRIVERNAME);
MODULE_LICENSE("GPL");

